"""
Routines for solving the nonlinear part of the Grad-Shafranov equation

Copyright 2016-2019 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from numpy import amin, amax


def solve(
    eq,
    profiles,
    constrain=None,
    rtol=1e-3,
    atol=1e-10,
    blend=0.0,
    show=False,
    axis=None,
    pause=0.0001,
    psi_bndry=None,
    maxits=50,
):
    """
    Perform Picard iteration to find solution to the Grad-Shafranov equation

    eq       - an Equilibrium object (equilibrium.py)
    profiles - A Profile object for toroidal current (jtor.py)

    rtol     - Relative tolerance (change in psi)/( max(psi) - min(psi) )
    atol     - Absolute tolerance, change in psi
    blend    - Blending of previous and next psi solution
               psi{n+1} <- psi{n+1} * (1-blend) + blend * psi{n}

    show     - If true, plot the plasma equilibrium at each nonlinear step
    axis     - Specify a figure to plot onto. Default (None) creates a new figure
    pause    - Delay between output plots. If negative, waits for window to be closed

    maxits   - Maximum number of iterations. Set to None for no limit.
               If this limit is exceeded then a RuntimeError is raised.
    """

    if constrain is not None:
        # Set the coil currents to get X-points in desired locations
        constrain(eq)

    # Get the total psi = plasma + coils
    psi = eq.psi()

    if show:
        import matplotlib.pyplot as plt
        from .plotting import plotEquilibrium
        from . import critical

        if pause > 0.0 and axis is None:
            # No axis specified, so create a new figure
            fig = plt.figure()
            axis = fig.add_subplot(111)

    iteration = 0  # Count number of iterations
    # Start main loop
    while True:
        if show:
            # Plot state of plasma equilibrium
            if pause < 0:
                fig = plt.figure()
                axis = fig.add_subplot(111)
            else:
                axis.clear()

            plotEquilibrium(eq, axis=axis, show=False)

            if pause < 0:
                # Wait for user to close the window
                plt.show()
            else:
                # Update the canvas and pause
                # Note, a short pause is needed to force drawing update
                axis.figure.canvas.draw()
                plt.pause(pause)

        # Copy psi to compare at the end
        psi_last = psi.copy()

        # Solve equilbrium, using the given psi to calculate Jtor
        eq.solve(profiles, psi=psi, psi_bndry=psi_bndry)

        # Get the new psi, including coils
        psi = eq.psi()

        # Compare against last solution
        psi_change = psi_last - psi
        psi_maxchange = amax(abs(psi_change))
        psi_relchange = psi_maxchange / (amax(psi) - amin(psi))

        # Check if the relative change in psi is small enough
        if (psi_maxchange < atol) or (psi_relchange < rtol):
            break

        # Adjust the coil currents
        if constrain is not None:
            constrain(eq)

        psi = (1.0 - blend) * eq.psi() + blend * psi_last

        # Check if the maximum iterations has been exceeded
        iteration += 1
        if maxits and iteration > maxits:
            raise RuntimeError(
                "Picard iteration failed to converge (too many iterations)"
            )



"""
Low level routines for reading and writing DivGeo files

Copyright 2018 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np

from ._fileutils import f2s, ChunkOutput, write_1d, write_2d, next_value


def write(data, fh, label=None):
    """
    Write a DivGeo file, given a dictionary of data

    data - dictionary
      nx, ny        Number of points in R (x), Z (y)
      rcentr        Reference value of R
      bcentr        Vacuum toroidal magnetic field at rcentr
      sibdry        Poloidal flux psi at plasma boundary

      either:
        r[nx]       1D array of R values
      or
        rdim        Size of the R dimension
        rleft       Innermost R point

      either:
        z[ny]       1D array of Z values
      or
        zdim        Size of the Z dimension
        zmid        Middle of the Z dimension

    """

    # Write header
    fh.write(
        """    jm   :=  no. of grid points in radial direction;
    km   :=  no. of grid points in vertical direction;
    r    :=  radial   coordinates of grid points  [m];
    z    :=  vertical coordinates of grid points  [m];
    psi  :=  flux per radiant at grid points     [Wb/rad];
    psib :=  psi at plasma boundary              [Wb/rad];
    btf  :=  toroidal magnetic field                  [t];
    rtf  :=  major radius at which btf is specified   [m];
 
 
    jm    =          {nx:d} ;
    km    =          {ny:d} ;
    psib  =   {sibdry:1.15E}  Wb/rad;
    btf   =   {bcentr:1.14f}       t;
    rtf   =   {rcentr:1.14f}       m;
 
""".format(
            **data
        )
    )

    try:
        r = data["r"]
    except KeyError:
        # No "r" in the dictionary
        # use rdim, rleft and nx (from eqdsk)

        Rmin = data["rleft"]
        Rmax = data["rleft"] + data["rdim"]
        nx = data["nx"]

        dR = (Rmax - Rmin) / (nx - 1)
        r = np.arange(nx) * dR + Rmin

    try:
        z = data["z"]
    except KeyError:
        # No "z" in the dictionary
        # use zdim, zmid and ny (from eqdsk)

        Zmin = data["zmid"] - 0.5 * data["zdim"]
        Zmax = data["zmid"] + 0.5 * data["zdim"]
        ny = data["ny"]

        dZ = (Zmax - Zmin) / (ny - 1)
        z = np.arange(ny) * dZ + Zmin

    # Now write r and z
    fh.write("    r(1:jm);\n")
    co = ChunkOutput(fh, chunksize=5, extraspaces=3)
    write_1d(r, co)
    co.newline()

    fh.write(" \n    z(1:km);\n")
    write_1d(z, co)
    co.newline()

    fh.write(" \n      ((psi(j,k)-psib,j=1,jm),k=1,km)\n")

    write_2d(data["psi"] - data["sibdry"], co)
    co.newline()



"""

fields - Lists the variables stored in the file, a default value, and a description

"""

from . import _fileutils as fu

import warnings

# List of file data variables, default values, and documentation
# This is used in both reader and writer
fields = [
    (
        "tsaisq",
        0.0,
        "total chi2 from magnetic probes, flux loops, Rogowski and external coils",
    ),
    ("rcencm", 100.0, "major radius in cm for vacuum field BCENTR"),
    ("bcentr", 1.0, "vacuum toroidal magnetic field in Tesla at RCENCM"),
    ("pasmat", 1e6, "measured plasma toroidal current in Ampere"),
    ("cpasma", 1e6, "fitted plasma toroidal current in Ampere-turn"),
    ("rout", 100.0, "major radius of geometric center in cm"),
    ("zout", 0.0, "Z of geometric center in cm"),
    ("aout", 50.0, "plasma minor radius in cm"),
    ("eout", 1.0, "Plasma boundary elongation"),
    ("doutu", 1.0, "upper triangularity"),
    ("doutl", 1.0, "lower triangularity"),
    ("vout", 1000.0, "plasma volume in cm3"),
    ("rcurrt", 100.0, "major radius in cm of current centroid"),
    ("zcurrt", 0.0, "Z in cm at current centroid"),
    ("qsta", 5.0, "equivalent safety factor q*"),
    ("betat", 1.0, "toroidal beta in %"),
    (
        "betap",
        1.0,
        "poloidal beta with normalization average poloidal magnetic BPOLAV defined through Ampere's law",
    ),
    (
        "ali",
        0.0,
        "li with normalization average poloidal magnetic defined through Ampere's law",
    ),
    ("oleft", 10.0, "plasma inner gap in cm"),
    ("oright", 10.0, "plasma outer gap in cm"),
    ("otop", 10.0, "plasma top gap in cm"),
    ("obott", 10.0, "plasma bottom gap in cm"),
    ("qpsib", 5.0, "q at 95% of poloidal flux"),
    ("vertn", 1.0, "vacuum field (index? -- seems to be float) at current centroid"),
    # fmt_1040 = r '^\s*' + 4 * r '([\s\-]\d+\.\d+[Ee][\+\-]\d\d)'
    # read(neqdsk, 1040)(rco2v(k, jj), k = 1, mco2v)
    (None, None, None),  # New line
    (
        "rco2v",
        lambda data: [0.0] * data["mco2v"],
        "1D array : path length in cm of vertical CO2 density chord",
    ),
    # read(neqdsk, 1040)(dco2v(jj, k), k = 1, mco2v)
    (None, None, None),  # New line
    (
        "dco2v",
        lambda data: [0.0] * data["mco2v"],
        "line average electron density in cm3 from vertical CO2 chord",
    ),
    # read(neqdsk, 1040)(rco2r(k, jj), k = 1, mco2r)
    (None, None, None),  # New line
    (
        "rco2r",
        lambda data: [0.0] * data["mco2r"],
        "path length in cm of radial CO2 density chord",
    ),
    # read(neqdsk, 1040)(dco2r(jj, k), k = 1, mco2r)
    (None, None, None),  # New line
    (
        "dco2r",
        lambda data: [0.0] * data["mco2r"],
        "line average electron density in cm3 from radial CO2 chord",
    ),
    (None, None, None),  # New line
    ("shearb", 0.0, ""),
    (
        "bpolav",
        1.0,
        "average poloidal magnetic field in Tesla defined through Ampere's law",
    ),
    ("s1", 0.0, "Shafranov boundary line integrals"),
    ("s2", 0.0, "Shafranov boundary line integrals"),
    ("s3", 0.0, "Shafranov boundary line integrals"),
    ("qout", 0.0, "q at plasma boundary"),
    ("olefs", 0.0, ""),
    ("orighs", 0.0, "outer gap of external second separatrix in cm"),
    ("otops", 0.0, "top gap of external second separatrix in cm"),
    ("sibdry", 1.0, ""),
    ("areao", 100.0, "cross sectional area in cm2"),
    ("wplasm", 0.0, ""),
    ("terror", 0.0, "equilibrium convergence error"),
    ("elongm", 0.0, "elongation at magnetic axis"),
    ("qqmagx", 0.0, "axial safety factor q(0)"),
    ("cdflux", 0.0, "computed diamagnetic flux in Volt-sec"),
    ("alpha", 0.0, "Shafranov boundary line integral parameter"),
    ("rttt", 0.0, "Shafranov boundary line integral parameter"),
    ("psiref", 1.0, "reference poloidal flux in VS/rad"),
    (
        "xndnt",
        0.0,
        "vertical stability parameter, vacuum field index normalized to critical index value",
    ),
    ("rseps1", 1.0, "major radius of x point in cm"),
    ("zseps1", -1.0, ""),
    ("rseps2", 1.0, "major radius of x point in cm"),
    ("zseps2", 1.0, ""),
    ("sepexp", 0.0, "separatrix radial expansion in cm"),
    ("obots", 0.0, "bottom gap of external second separatrix in cm"),
    ("btaxp", 1.0, "toroidal magnetic field at magnetic axis in Tesla"),
    ("btaxv", 1.0, "vacuum toroidal magnetic field at magnetic axis in Tesla"),
    ("aaq1", 100.0, "minor radius of q=1 surface in cm, 100 if not found"),
    ("aaq2", 100.0, "minor radius of q=2 surface in cm, 100 if not found"),
    ("aaq3", 100.0, "minor radius of q=3 surface in cm, 100 if not found"),
    (
        "seplim",
        0.0,
        "> 0 for minimum gap in cm in divertor configurations, < 0 absolute value for minimum distance to external separatrix in limiter configurations",
    ),
    ("rmagx", 100.0, "major radius in cm at magnetic axis"),
    ("zmagx", 0.0, ""),
    ("simagx", 0.0, "Poloidal flux at the magnetic axis"),
    ("taumhd", 0.0, "energy confinement time in ms"),
    ("betapd", 0.0, "diamagnetic poloidal b"),
    ("betatd", 0.0, "diamagnetic toroidal b in %"),
    ("wplasmd", 0.0, "diamagnetic plasma stored energy in Joule"),
    ("diamag", 0.0, "measured diamagnetic flux in Volt-sec"),
    ("vloopt", 0.0, "measured loop voltage in volt"),
    ("taudia", 0.0, "diamagnetic energy confinement time in ms"),
    (
        "qmerci",
        0.0,
        "Mercier stability criterion on axial q(0), q(0) > QMERCI for stability",
    ),
    ("tavem", 0.0, "average time in ms for magnetic and MSE data"),
    # ishot > 91000
    # The next section is dependent on the EFIT version
    # New version of EFIT on 05/24/97 writes aeqdsk that includes
    # data values for parameters nsilop,magpri,nfcoil and nesum.
    (None, True, None),  # New line
    (
        "nsilop",
        lambda data: len(data.get("csilop", [])),
        "Number of flux loop signals, len(csilop)",
    ),
    (
        "magpri",
        lambda data: len(data.get("cmpr2", [])),
        "Number of flux loop signals, len(cmpr2) (added to nsilop)",
    ),
    (
        "nfcoil",
        lambda data: len(data.get("ccbrsp", [])),
        "Number of calculated external coil currents, len(ccbrsp)",
    ),
    (
        "nesum",
        lambda data: len(data.get("eccurt", [])),
        "Number of measured E-coil currents",
    ),
    (None, None, None),  # New line
    (
        "csilop",
        lambda data: [0.0] * data.get("nsilop", 0),
        "computed flux loop signals in Weber",
    ),
    ("cmpr2", lambda data: [0.0] * data.get("magpri", 0), ""),
    (
        "ccbrsp",
        lambda data: [0.0] * data.get("nfcoil", 0),
        "computed external coil currents in Ampere",
    ),
    (
        "eccurt",
        lambda data: [0.0] * data.get("nesum", 0),
        "measured E-coil current in Ampere",
    ),
    ("pbinj", 0.0, "neutral beam injection power in Watts"),
    ("rvsin", 0.0, "major radius of vessel inner hit spot in cm"),
    ("zvsin", 0.0, "Z of vessel inner hit spot in cm"),
    ("rvsout", 0.0, "major radius of vessel outer hit spot in cm"),
    ("zvsout", 0.0, "Z of vessel outer hit spot in cm"),
    ("vsurfa", 0.0, "plasma surface loop voltage in volt, E EQDSK only"),
    ("wpdot", 0.0, "time derivative of plasma stored energy in Watt, E EQDSK only"),
    ("wbdot", 0.0, "time derivative of poloidal magnetic energy in Watt, E EQDSK only"),
    ("slantu", 0.0, ""),
    ("slantl", 0.0, ""),
    ("zuperts", 0.0, ""),
    ("chipre", 0.0, "total chi2 pressure"),
    ("cjor95", 0.0, ""),
    ("pp95", 0.0, "normalized P'(y) at 95% normalized poloidal flux"),
    ("ssep", 0.0, ""),
    ("yyy2", 0.0, "Shafranov Y2 current moment"),
    ("xnnc", 0.0, ""),
    ("cprof", 0.0, "current profile parametrization parameter"),
    ("oring", 0.0, "not used"),
    (
        "cjor0",
        0.0,
        "normalized flux surface average current density at 99% of normalized poloidal flux",
    ),
    ("fexpan", 0.0, "flux expansion at x point"),
    ("qqmin", 0.0, "minimum safety factor qmin"),
    ("chigamt", 0.0, "total chi2 MSE"),
    ("ssi01", 0.0, "magnetic shear at 1% of normalized poloidal flux"),
    ("fexpvs", 0.0, "flux expansion at outer lower vessel hit spot"),
    (
        "sepnose",
        0.0,
        "radial distance in cm between x point and external field line at ZNOSE",
    ),
    ("ssi95", 0.0, "magnetic shear at 95% of normalized poloidal flux"),
    ("rqqmin", 0.0, "normalized radius of qmin , square root of normalized volume"),
    ("cjor99", 0.0, ""),
    (
        "cj1ave",
        0.0,
        "normalized average current density in plasma outer 5% normalized poloidal flux region",
    ),
    ("rmidin", 0.0, "inner major radius in m at Z=0.0"),
    ("rmidout", 0.0, "outer major radius in m at Z=0.0"),
]


def write(data, fh):
    """
    data  [dict] - keys are given with documentation in the `fields` list.
        Also includes
         shot [int] - The shot number
    time  - in ms

    """
    # First line identification string
    # Default to date > 1997 since that format includes nsilop etc.
    fh.write("{0:11s}\n".format(data.get("header", " 26-OCT-98 09/07/98  ")))

    # Second line shot number
    fh.write(" {:d}               1\n".format(data.get("shot", 0)))

    # Third line time
    fh.write(" " + fu.f2s(data.get("time", 0.0)) + "\n")

    # Fourth line
    # time(jj),jflag(jj),lflag,limloc(jj), mco2v,mco2r,qmflag
    #   jflag = 0 if error  (? Seems to contradict example)
    #   lflag > 0 if error  (? Seems to contradict example)
    #   limloc  IN/OUT/TOP/BOT: limiter inside/outside/top/bot SNT/SNB: single null top/bottom DN: double null
    #   mco2v   number of vertical CO2 density chords
    #   mco2r   number of radial CO2 density chords
    #   qmflag  axial q(0) flag, FIX if constrained and CLC for float
    fh.write(
        "*{:s}             {:d}                {:d} {:s}  {:d}   {:d} {:s}\n".format(
            fu.f2s(data.get("time", 0.0)).strip(),
            data.get("jflag", 1),
            data.get("lflag", 0),
            data.get("limloc", "DN"),
            data.get("mco2v", 0),
            data.get("mco2r", 0),
            data.get("qmflag", "CLC"),
        )
    )
    # Output data in lines of 4 values each
    with fu.ChunkOutput(fh, chunksize=4) as output:
        for key, default, description in fields:
            if callable(default):
                # Replace the default function with the value, which may depend on previously read data
                default = default(data)

            if key is None:
                output.newline()  # Ensure on a new line
            else:
                output.write(data.get(key, default))


def read(fh):
    """
    Read an AEQDSK file, returning a dictionary of data
    """
    # First line label. Date.
    header = fh.readline()

    # Second line shot number
    shot = int(fh.readline().split()[0])

    # Third line time [ms]
    time = float(fh.readline())

    # Fourth line has (up to?) 9 entries
    # time(jj),jflag(jj),lflag,limloc(jj), mco2v,mco2r,qmflag
    words = fh.readline().split()

    # Dictionary to hold result
    data = {
        "header": header,
        "shot": shot,
        "time": time,
        "jflag": int(words[1]),
        "lflag": int(words[2]),
        "limloc": words[3],  # e.g. "SNB"
        "mco2v": int(words[4]),
        "mco2r": int(words[5]),
        "qmflag": words[6],
    }  # e.g. "CLC"

    # Read each value from the file, and put into variables
    values = fu.next_value(fh)
    for key, default, doc in fields:
        if key is None:
            continue  # skip

        if callable(default):
            default = default(data)

        if isinstance(default, list):
            # Read a list the same length as the default
            data[key] = [next(values) for elt in default]
        else:
            value = next(values)
            if isinstance(default, int) and not isinstance(value, int):
                # Expecting an integer, but didn't get one
                warnings.warn("Expecting an integer for '" + key + "' in aeqdsk file")
                break
            data[key] = value

    return data



from setuptools import setup
import os

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

import freegs

setup(
    name="FreeGS",
    version=freegs.__version__,
    packages=["freegs"],

    license="LGPL",
    author="Ben Dudson",
    author_email='benjamin.dudson@york.ac.uk',
    url="https://github.com/bendudson/freegs",
    description="Free boundary Grad-Shafranov solver for tokamak plasma equilibria",

    long_description=read("README.md"),
    
    install_requires=['numpy>=1.8',
                      'scipy>=0.14',
                      'matplotlib>=1.3',
                      'h5py==2.10.0'],
    
    platforms='any',

    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Physics'
        ],
)



"""
Grad-Shafranov equation

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from numpy import zeros

# Elliptic integrals of first and second kind (K and E)
from scipy.special import ellipk, ellipe

from numpy import sqrt, pi, clip

from scipy.sparse import lil_matrix, eye

# Physical constants
mu0 = 4e-7 * pi


class GSElliptic:
    """
    Represents the Grad-Shafranov elliptic operator

    \Delta^* = R^2 \nabla\cdot\frac{1}{R^2}\nabla

    which is

    d^2/dR^2 + d^2/dZ^2 - (1/R)*d/dR

    """

    def __init__(self, Rmin):
        """
        Specify minimum radius

        """
        self.Rmin = Rmin

    def __call__(self, psi, dR, dZ):
        """
        Apply the full operator to 2D field f

        Inputs
        ------

        psi[R,Z]   A 2D NumPy array containing flux function psi
        dR         A scalar (uniform grid spacing)
        dZ         A scalar (uniform grid spacing)

        """
        nx = psi.shape[0]
        ny = psi.shape[1]

        b = zeros([nx, ny])

        invdR2 = 1.0 / dR ** 2
        invdZ2 = 1.0 / dZ ** 2

        for x in range(1, nx - 1):
            R = self.Rmin + dR * x  # Major radius of this point
            for y in range(1, ny - 1):
                # Loop over points in the domain
                b[x, y] = (
                    psi[x, y - 1] * invdZ2
                    + (invdR2 + 1.0 / (2.0 * R * dR)) * psi[x - 1, y]
                    - 2.0 * (invdR2 + invdZ2) * psi[x, y]
                    + (invdR2 - 1.0 / (2.0 * R * dR)) * psi[x + 1, y]
                    + psi[x, y + 1] * invdZ2
                )
        return b

    def diag(self, dR, dZ):
        """
        Return the diagonal elements

        """
        return -2.0 / dR ** 2 - 2.0 / dZ ** 2


class GSsparse:
    """
    Calculates sparse matrices for the Grad-Shafranov operator
    """

    def __init__(self, Rmin, Rmax, Zmin, Zmax):
        self.Rmin = Rmin
        self.Rmax = Rmax
        self.Zmin = Zmin
        self.Zmax = Zmax

    def __call__(self, nx, ny):
        """
        Create a sparse matrix with given resolution

        """

        # Calculate grid spacing
        dR = (self.Rmax - self.Rmin) / (nx - 1)
        dZ = (self.Zmax - self.Zmin) / (ny - 1)

        # Total number of points
        N = nx * ny

        # Create a linked list sparse matrix
        A = eye(N, format="lil")

        invdR2 = 1.0 / dR ** 2
        invdZ2 = 1.0 / dZ ** 2

        for x in range(1, nx - 1):
            R = self.Rmin + dR * x  # Major radius of this point
            for y in range(1, ny - 1):
                # Loop over points in the domain
                row = x * ny + y

                # y-1
                A[row, row - 1] = invdZ2

                # x-1
                A[row, row - ny] = invdR2 + 1.0 / (2.0 * R * dR)

                # diagonal
                A[row, row] = -2.0 * (invdR2 + invdZ2)

                # x+1
                A[row, row + ny] = invdR2 - 1.0 / (2.0 * R * dR)

                # y+1
                A[row, row + 1] = invdZ2

        # Convert to Compressed Sparse Row (CSR) format
        return A.tocsr()


class GSsparse4thOrder:
    """
    Calculates sparse matrices for the Grad-Shafranov operator using 4th-order
    finite differences.
    """

    # Coefficients for first derivatives
    # (index offset, weight)

    centred_1st = [(-2, 1.0 / 12), (-1, -8.0 / 12), (1, 8.0 / 12), (2, -1.0 / 12)]

    offset_1st = [
        (-1, -3.0 / 12),
        (0, -10.0 / 12),
        (1, 18.0 / 12),
        (2, -6.0 / 12),
        (3, 1.0 / 12),
    ]

    # Coefficients for second derivatives
    # (index offset, weight)
    centred_2nd = [
        (-2, -1.0 / 12),
        (-1, 16.0 / 12),
        (0, -30.0 / 12),
        (1, 16.0 / 12),
        (2, -1.0 / 12),
    ]

    offset_2nd = [
        (-1, 10.0 / 12),
        (0, -15.0 / 12),
        (1, -4.0 / 12),
        (2, 14.0 / 12),
        (3, -6.0 / 12),
        (4, 1.0 / 12),
    ]

    def __init__(self, Rmin, Rmax, Zmin, Zmax):
        self.Rmin = Rmin
        self.Rmax = Rmax
        self.Zmin = Zmin
        self.Zmax = Zmax

    def __call__(self, nx, ny):
        """
        Create a sparse matrix with given resolution

        """

        # Calculate grid spacing
        dR = (self.Rmax - self.Rmin) / (nx - 1)
        dZ = (self.Zmax - self.Zmin) / (ny - 1)

        # Total number of points, including boundaries
        N = nx * ny

        # Create a linked list sparse matrix
        A = lil_matrix((N, N))

        invdR2 = 1.0 / dR ** 2
        invdZ2 = 1.0 / dZ ** 2

        for x in range(1, nx - 1):
            R = self.Rmin + dR * x  # Major radius of this point
            for y in range(1, ny - 1):
                row = x * ny + y

                # d^2 / dZ^2
                if y == 1:
                    # One-sided derivatives in Z
                    for offset, weight in self.offset_2nd:
                        A[row, row + offset] += weight * invdZ2
                elif y == ny - 2:
                    # One-sided, reversed direction.
                    # Note that for second derivatives the sign of the weights doesn't change
                    for offset, weight in self.offset_2nd:
                        A[row, row - offset] += weight * invdZ2
                else:
                    # Central differencing
                    for offset, weight in self.centred_2nd:
                        A[row, row + offset] += weight * invdZ2

                # d^2 / dR^2 - (1/R) d/dR

                if x == 1:
                    for offset, weight in self.offset_2nd:
                        A[row, row + offset * ny] += weight * invdR2

                    for offset, weight in self.offset_1st:
                        A[row, row + offset * ny] -= weight / (R * dR)

                elif x == nx - 2:
                    for offset, weight in self.offset_2nd:
                        A[row, row - offset * ny] += weight * invdR2

                    for offset, weight in self.offset_1st:
                        A[row, row - offset * ny] += weight / (R * dR)
                else:
                    for offset, weight in self.centred_2nd:
                        A[row, row + offset * ny] += weight * invdR2

                    for offset, weight in self.centred_1st:
                        A[row, row + offset * ny] -= weight / (R * dR)

        # Set boundary rows
        for x in range(nx):
            for y in [0, ny - 1]:
                row = x * ny + y
                A[row, row] = 1.0
        for x in [0, nx - 1]:
            for y in range(ny):
                row = x * ny + y
                A[row, row] = 1.0

        # Convert to Compressed Sparse Row (CSR) format
        return A.tocsr()


def Greens(Rc, Zc, R, Z):
    """
    Calculate poloidal flux at (R,Z) due to a unit current
    at (Rc,Zc) using Greens function

    """

    # Calculate k^2
    k2 = 4.0 * R * Rc / ((R + Rc) ** 2 + (Z - Zc) ** 2)

    # Clip to between 0 and 1 to avoid nans e.g. when coil is on grid point
    k2 = clip(k2, 1e-10, 1.0 - 1e-10)
    k = sqrt(k2)

    # Note definition of ellipk, ellipe in scipy is K(k^2), E(k^2)
    return (
        (mu0 / (2.0 * pi))
        * sqrt(R * Rc)
        * ((2.0 - k2) * ellipk(k2) - 2.0 * ellipe(k2))
        / k
    )


def GreensBz(Rc, Zc, R, Z, eps=1e-3):
    """
    Calculate vertical magnetic field at (R,Z)
    due to unit current at (Rc, Zc)

    Bz = (1/R) d psi/dR
    """

    return (Greens(Rc, Zc, R + eps, Z) - Greens(Rc, Zc, R - eps, Z)) / (2.0 * eps * R)


def GreensBr(Rc, Zc, R, Z, eps=1e-3):
    """
    Calculate radial magnetic field at (R,Z)
    due to unit current at (Rc, Zc)

    Br = -(1/R) d psi/dZ
    """

    return (Greens(Rc, Zc, R, Z - eps) - Greens(Rc, Zc, R, Z + eps)) / (2.0 * eps * R)



"""
Routines for optimising equilibria
These make use of the generic "optimiser" routines

Copyright 2019 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from . import optimiser
from . import polygons
from . import picard

from math import sqrt

### Measures which operate on Equilibrium objects


def max_abs_coil_current(eq):
    """
    Given an equilibrium, return the maximum absolute coil current
    """
    currents = eq.tokamak.getCurrents()  # dictionary
    return max(abs(current) for current in currents.values())


def max_coil_force(eq):
    forces = eq.tokamak.getForces()  # dictionary
    return max(sqrt(force[0] ** 2 + force[1] ** 2) for force in forces.values())


def no_wall_intersection(eq):
    """Prevent intersection of LCFS with walls by returning inf if intersections are found"""
    if eq.intersectsWall():
        return float("inf")
    return 0.0  # No intersection


### Combine measures


def weighted_sum(*args):
    """
    Combine measures together

    Returns a function which takes a single argument (the equilibrium),
    and passes it to a set of given functions. These functions are assumed
    to return values, which are multiplied by weights, summed and returned.

    args should be either functions or pairs of functions and weights.
    If no weights are supplied then a weight of 1.0 is assumed.
    """
    args_with_weights = []
    for arg in args:
        if callable(arg):
            args_with_weights.append((arg, 1.0))
        else:
            args_with_weights.append(arg)

    return lambda eq: sum(func(eq) * weight for func, weight in args_with_weights)


#### Controls for Equilibrium objects


class CoilRadius:
    """A control to modify the radius of a specified coil"""

    def __init__(self, label, minimum=0.0, maximum=None):
        """
        label  : string   The label of a coil to be modified
        minimum : number, optional   The minimum allowed radius
        maximum : number, optional   The maximum allowed radius
        """
        self.label = label
        self.minimum = minimum
        self.maximum = maximum

    def set(self, eq, R):
        if self.minimum and (R < self.minimum):
            R = self.minimum
        if self.maximum and (R > self.maximum):
            R = self.maximum
        eq.tokamak[self.label].R = R

    def get(self, eq):
        return eq.tokamak[self.label].R


class CoilHeight:
    """A control to modify the height of a specified coil"""

    def __init__(self, label, minimum=None, maximum=None):
        """
        label  : string   The label of a coil to be modified
        minimum : number, optional   The minimum allowed height
        maximum : number, optional   The maximum allowed height
        """
        self.label = label
        self.minimum = minimum
        self.maximum = maximum

    def set(self, eq, Z):
        if self.minimum and (Z < self.minimum):
            Z = self.minimum
        if self.maximum and (Z > self.maximum):
            Z = self.maximum
        eq.tokamak[self.label].Z = Z

    def get(self, eq):
        return eq.tokamak[self.label].Z


import matplotlib.pyplot as plt
from freegs.plotting import plotEquilibrium

#### Monitor optimisation solutions

# Plot and save the best equilibrium each generation
class PlotMonitor:
    """
    Plot the best solution at the end of each generation,
    saves the plot to a PNG file.
    """

    def __init__(self):
        self.fig = plt.figure()
        self.axis = self.fig.add_subplot(111)

    def __call__(self, generation, best, population):
        self.axis.clear()
        plotEquilibrium(best[1], axis=self.axis, show=False)
        # Update the canvas and pause
        # Note, a short pause is needed to force drawing update
        self.fig.canvas.draw()
        self.axis.set_title("Generation: {} Score: {}".format(generation, best[0]))
        self.fig.savefig("generation_{}.pdf".format(generation))
        plt.pause(0.5)


def optimise(eq, controls, measure, maxgen=10, N=10, CR=0.3, F=1.0, monitor=None):
    """Use Differential Evolution to optimise an Equilibrium
    https://en.wikipedia.org/wiki/Differential_evolution

    eq   is an Equilibrium to be used as starting solution
         These are deep copied, and passed as arguments
         to controls and measure functions

    controls   A list of control objects. These objects must
               have methods:
         .set(eq, value) which modifies the given Equilibrium
         .get(eq)  which returns the value from the Equilibrium

    measure(eq) is a function which returns a score (value) for a
                given Equilibrium. The optimiser tries to minimise this
                value.

    maxgen   is the maximum number of generations

    N >= 4 is the population size
    CR [0,1] is the crossover probability
    F  [0,2] is the differential weight

    monitor(generation, best, population)

           A function to be called each generation with the best
           Equilibrium and the whole population
           generation = integer
           best = (score, object)
           population = [(score, object)]

    Returns the Equilibrium with the lowest measure (score).

    """

    def solve_and_measure(eq):
        # Need to update some internal caches
        eq._pgreen = eq.tokamak.createPsiGreens(eq.R, eq.Z)
        try:
            # Re-solve
            picard.solve(eq, eq._profiles, eq._constraints)
        except:
            # Solve failed.
            return float("inf")
        # Call user-supplied evaluation function
        return measure(eq)

    # Call the generic optimiser,
    return optimiser.optimise(
        eq, controls, solve_and_measure, maxgen=maxgen, N=N, CR=CR, F=F, monitor=monitor
    )



"""
Class for reading/writing freegs `Equilibrium` objects

Currently just HDF5 via h5py

License
-------

Copyright 2018 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

# Py2/3 compatibility: h5py needs unicode for group/dataset names
from __future__ import unicode_literals

try:
    import h5py

    has_hdf5 = True
except ImportError:
    has_hdf5 = False

import numpy as np

from .equilibrium import Equilibrium
from .machine import Coil, Circuit, Solenoid, Wall, Machine
from . import boundary
from . import machine


class OutputFormatNotAvailableError(Exception):
    """Raised when we couldn't import HDF5 (or some other library)
    required for this OutputFile format

    """

    def __init__(self, file_format="HDF5"):
        self.message = "Sorry, {} is not available!".format(file_format)


class OutputFile(object):
    """
    Read/write freegs Equilibrium objects to file

    Currently supports HDF5 format only

    Given an Equilibrium object, eq, write to file like:

    >>> with freegs.OutputFile("test_readwrite.h5", 'w') as f:
    ...     f.write_equilibrium(eq)

    Read back into an Equilibrium like so:

    >>> with freegs.OutputFile("test_readwrite.h5", 'r') as f:
    ...      eq = f.read_equilibrium()

    Parameters
    ----------
    name : str
           Name of file to open
    mode : str
           Mode string to pass to `h5py.File`, one of:
             r  - Read-only, file must exist
             r+ - Read-write, file must exist
             w  - Create file, truncate if exists
             w- - Create file, fail if exists
             a  - Read-write if exists, create otherwise
    **kwds
           Other keyword arguments to pass to `h5py.File`
    """

    # Names of the groups in the file
    EQUILIBRIUM_GROUP_NAME = "equilbrium"
    MACHINE_GROUP_NAME = "tokamak"
    COILS_GROUP_NAME = "coils"

    def __init__(self, name, mode=None, **kwds):
        if not has_hdf5:
            raise OutputFormatNotAvailableError("HDF5")

        self.handle = h5py.File(name, mode, **kwds)

    def close(self):
        """
        Close the file
        """
        self.handle.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def write_equilibrium(self, equilibrium):
        """
        Write `equilbrium` to file
        """

        self.handle["coil_dtype"] = Coil.dtype
        self.handle["circuit_dtype"] = Circuit.dtype
        self.handle["solenoid_dtype"] = Solenoid.dtype

        type_to_dtype = {
            Coil.dtype: self.handle["coil_dtype"],
            Circuit.dtype: self.handle["circuit_dtype"],
            Solenoid.dtype: self.handle["solenoid_dtype"],
        }

        equilibrium_group = self.handle.require_group(self.EQUILIBRIUM_GROUP_NAME)

        equilibrium_group.create_dataset("Rmin", data=equilibrium.Rmin)
        equilibrium_group.create_dataset("Rmax", data=equilibrium.Rmax)
        equilibrium_group.create_dataset("R_1D", data=equilibrium.R_1D)
        equilibrium_group.create_dataset("R", data=equilibrium.R)

        equilibrium_group.create_dataset("Zmin", data=equilibrium.Zmin)
        equilibrium_group.create_dataset("Zmax", data=equilibrium.Zmax)
        equilibrium_group.create_dataset("Z_1D", data=equilibrium.Z_1D)
        equilibrium_group.create_dataset("Z", data=equilibrium.Z)

        equilibrium_group.create_dataset("current", data=equilibrium.plasmaCurrent())
        equilibrium_group["current"].attrs["title"] = "Plasma current [Amps]"

        psi_id = equilibrium_group.create_dataset("psi", data=equilibrium.psi())
        psi_id.dims[0].label = "R"
        psi_id.dims[1].label = "Z"
        psi_id.dims.create_scale(equilibrium_group["R_1D"], "R")
        psi_id.dims.create_scale(equilibrium_group["Z_1D"], "Z")
        psi_id.dims[0].attach_scale(equilibrium_group["R_1D"])
        psi_id.dims[1].attach_scale(equilibrium_group["Z_1D"])

        plasma_psi_id = equilibrium_group.create_dataset(
            "plasma_psi", data=equilibrium.plasma_psi
        )
        plasma_psi_id.dims[0].label = "R"
        plasma_psi_id.dims[1].label = "Z"
        plasma_psi_id.dims[0].attach_scale(equilibrium_group["R_1D"])
        plasma_psi_id.dims[1].attach_scale(equilibrium_group["Z_1D"])

        equilibrium_group.create_dataset(
            "boundary_function", data=equilibrium._applyBoundary.__name__
        )

        tokamak_group = equilibrium_group.create_group(self.MACHINE_GROUP_NAME)

        if equilibrium.tokamak.wall is not None:
            tokamak_group.create_dataset("wall_R", data=equilibrium.tokamak.wall.R)
            tokamak_group.create_dataset("wall_Z", data=equilibrium.tokamak.wall.Z)

        coils_group = tokamak_group.create_group(self.COILS_GROUP_NAME)
        for label, coil in equilibrium.tokamak.coils:
            dtype = type_to_dtype[coil.dtype]
            coils_group.create_dataset(
                label, dtype=dtype, data=np.array(coil.to_numpy_array())
            )
            # A bit gross, but store the class name so we know what
            # type to restore it to later
            coils_group[label].attrs["freegs type"] = coil.__class__.__name__

    def read_equilibrium(self):
        """
        Read an equilibrium from the file

        Returns
        -------
        Equilibrium
            A new `Equilibrium` object
        """

        equilibrium_group = self.handle[self.EQUILIBRIUM_GROUP_NAME]
        tokamak_group = equilibrium_group[self.MACHINE_GROUP_NAME]
        coil_group = tokamak_group[self.COILS_GROUP_NAME]

        # This is also a bit hacky - find the appropriate class in
        # freegs.machine and then call the `from_numpy_array` class
        # method
        def make_coil_set(thing):
            return machine.__dict__[thing.attrs["freegs type"]].from_numpy_array(thing)

        # Unfortunately this creates the coils in lexographical order
        # by label, losing the origin
        coils = [(label, make_coil_set(coil)) for label, coil in coil_group.items()]

        if "wall_R" in tokamak_group:
            wall_R = tokamak_group["wall_R"][:]
            wall_Z = tokamak_group["wall_Z"][:]
            wall = Wall(wall_R, wall_Z)
        else:
            wall = None

        tokamak = Machine(coils, wall)

        Rmin = equilibrium_group["Rmin"][()]
        Rmax = equilibrium_group["Rmax"][()]
        Zmin = equilibrium_group["Zmin"][()]
        Zmax = equilibrium_group["Zmax"][()]
        nx, ny = equilibrium_group["R"].shape

        current = equilibrium_group["current"][()]
        plasma_psi = equilibrium_group["plasma_psi"][()]

        # Feels a bit hacky... the boundary function is saved as a
        # string of the function __name__, which we then look up in
        # the boundary module dict
        eq_boundary_name = equilibrium_group["boundary_function"][()]
        eq_boundary_func = boundary.__dict__[eq_boundary_name]

        equilibrium = Equilibrium(
            tokamak=tokamak,
            Rmin=Rmin,
            Rmax=Rmax,
            Zmin=Zmin,
            Zmax=Zmax,
            nx=nx,
            ny=ny,
            psi=plasma_psi,
            current=current,
            boundary=eq_boundary_func,
        )

        return equilibrium



"""
Quadrature rules for averaging over polygons

Note: integration weights set so that sum of weights is 1 and gives
the average of a function over the polygon, not the integral.

License
-------

Copyright 2019 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from . import polygons


def triangle_quad(triangle, n=6):
    """
    Given a triangle in the form [(r1,z1), (r2,z2), (r3,z3)]
    returns a list of evaluation points [(r,z,weight),...]

    n   number of quadrature points.
        currently: 1, 3 or 6

    Coefficients taken from http://www.cs.rpi.edu/~flaherje/pdf/fea6.pdf
    Joseph E. Flaherty course notes, Rensselaer Polytechnic Institute
    """

    assert len(triangle) == 3

    r1, z1 = triangle[0]
    r2, z2 = triangle[1]
    r3, z3 = triangle[2]

    if n == 1:
        # One point in the middle of the triangle
        return [((r1 + r2 + r3) / 3, (z1 + z2 + z3) / 3, 1.0)]

    elif n == 3:
        return [
            ((4 * r1 + r2 + r3) / 6, (4 * z1 + z2 + z3) / 6, 1.0 / 3),
            ((r1 + 4 * r2 + r3) / 6, (z1 + 4 * z2 + z3) / 6, 1.0 / 3),
            ((r1 + r2 + 4 * r3) / 6, (z1 + z2 + 4 * z3) / 6, 1.0 / 3),
        ]

    elif n == 6:
        a = 0.816847572980459
        b = 0.5 * (1.0 - a)

        c = 0.108103018168070
        d = 0.5 * (1.0 - c)

        return [
            ((a * r1 + b * r2 + b * r3), (a * z1 + b * z2 + b * z3), 0.109951743655322),
            ((b * r1 + a * r2 + b * r3), (b * z1 + a * z2 + b * z3), 0.109951743655322),
            ((b * r1 + b * r2 + a * r3), (b * z1 + b * z2 + a * z3), 0.109951743655322),
            ((c * r1 + d * r2 + d * r3), (c * z1 + d * z2 + d * z3), 0.223381589678011),
            ((d * r1 + c * r2 + d * r3), (d * z1 + c * z2 + d * z3), 0.223381589678011),
            ((d * r1 + d * r2 + c * r3), (d * z1 + d * z2 + c * z3), 0.223381589678011),
        ]
    else:
        raise ValueError("Quadrature not available for n={}".format(n))


def polygon_quad(polygon, n=6):
    """
    Given a polygon in the form [(r1,z1), (r2,z2), (r3,z3), ...]
    calculates a set of quadrature points and weights, by splitting
    the polygon into triangles.

    returns a list of evaluation points and weights [(r,z,weight),...]
    These are normalised to calculate the average value of a function
    over the polygon; multiply by the area to get the integral.

    n   number of quadrature points in each triangle
        currently: 1, 3 or 6
    """

    # Split polygon into triangles
    triangles = polygons.triangulate(polygon)

    # Calculate the area of each triangle, to get a relative weighting
    areas = [polygons.area(triangle) for triangle in triangles]
    total_area = sum(areas)

    quadrature = []  # List of all points
    for triangle, area in zip(triangles, areas):
        points = triangle_quad(triangle, n=n)  # Quadrature points for this triangle
        quadrature += [
            (r, z, w * area / total_area) for r, z, w in points
        ]  # Modify the weights

    return quadrature


def average(func, quad):
    """
    Average func(r,z) using given quadrature
    """
    return sum(func(r, z) * w for r, z, w in quad)
